/*
 * PROJECT: napkit at http://matereal.sourceforge.jp/
 * ----------------------------------------------------------------------------
 *
 * This file is part of NyARToolkit Application Toolkit.
 * NyARToolkit Application Toolkit, or simply "napkit",
 * is a simple wrapper library for NyARToolkit.
 *
 * ----------------------------------------------------------------------------
 *
 * License version: GPL 3.0
 *
 * napkit is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * napkit is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with napkit. If not, see <http://www.gnu.org/licenses/>.
 */
package jp.digitalmuseum.napkit;

import java.util.HashSet;
import java.util.Set;

import jp.digitalmuseum.utils.ScreenRectangle;
import jp.digitalmuseum.utils.VariableLengthArray;
import jp.nyatla.nyartoolkit.NyARException;
import jp.nyatla.nyartoolkit.core.match.NyARMatchPattDeviationColorData;
import jp.nyatla.nyartoolkit.core.pickup.INyARColorPatt;
import jp.nyatla.nyartoolkit.core.pickup.NyARColorPatt_O3;
import jp.nyatla.nyartoolkit.core.squaredetect.NyARSquare;
import jp.nyatla.nyartoolkit.core.squaredetect.NyARSquareStack;


/**
 * Marker detector class using NyARToolKit.
 *
 * @author Jun KATO
 * @see NapMarkerDetector
 */
public class NapMarkerDetectorImpl extends NapSquareDetectorImpl implements NapMarkerDetector {
	final private VariableLengthArray<NapDetectionResult> resultHolder
			= new VariableLengthArray<NapDetectionResult>();

	private HashSet<NapMarker> markers;
	private INyARColorPatt squareImage;
	private NyARMatchPattDeviationColorData deviationData;

	private NapMarkerMatchingStrategy matchingStrategy;

	/** Constructor. */
	public NapMarkerDetectorImpl() {
		super();
		markers = new HashSet<NapMarker>();
		matchingStrategy = new NapSimpleMMStrategy();
	}

	/**
	 * Update marker size and initialize workspaces for comparing a marker with images.
	 */
	private void updateMarkerSize() {
		final NapMarker sampleMarker = markers.iterator().next();
		final int
				markerWidth = sampleMarker.getWidth(),
				markerHeight = sampleMarker.getHeight();
		squareImage = new NyARColorPatt_O3(markerWidth, markerHeight);
		deviationData = new NyARMatchPattDeviationColorData(markerWidth, markerHeight);
	}

	public boolean addMarkers(Set<NapMarker> markers) {
		boolean needUpdateMarkerSize = true;
		int markerWidth = -1, markerHeight = -1;
		if (markers.size() > 0) {
			final NapMarker sampleMarker = markers.iterator().next();
			markerWidth = sampleMarker.getWidth();
			markerHeight = sampleMarker.getHeight();
			needUpdateMarkerSize = false;
		}
		for (NapMarker marker : markers) {
			if (markerWidth == -1
					&& markerHeight == -1) {
				markerWidth = marker.getWidth();
				markerHeight = marker.getHeight();
			} else {
				if (markerWidth != marker.getWidth()
						|| markerHeight != marker.getHeight()) {
					return false;
				}
			}
			this.markers.add(marker);
		}
		if (needUpdateMarkerSize) {
			updateMarkerSize();
		}
		return true;
	}

	public boolean addMarker(NapMarker marker) {
		if (markers.size() == 0) {
			markers.add(marker);
			updateMarkerSize();
			return true;
		}
		final NapMarker sampleMarker = markers.iterator().next();
		if (sampleMarker.getWidth() != marker.getWidth()
				|| sampleMarker.getHeight() != marker.getHeight()) {
			return false;
		}
		markers.add(marker);
		return true;
	}

	public void setMatchingStrategy(NapMarkerMatchingStrategy matchingStrategy) {
		this.matchingStrategy = matchingStrategy;
	}

	public VariableLengthArray<NapDetectionResult> detectMarker(byte[] imageData) {
		image.setData(imageData);
		final NyARSquareStack squareStack = detectRawSquare(image);
		final int squaresFound = squareStack.getLength();
		resultHolder.clear();
		resultHolder.realloc(squaresFound);
		if (squaresFound < 1) {
			return resultHolder;
		}

		if (markers == null) {
			throw new IllegalStateException("No markers to detect.");
		}

		for (int i = 0; i < squaresFound; i++) {

			// Pick a square image.
			final NyARSquare square = (NyARSquare) squareStack.getItem(i);
			boolean pick;
			try {
				pick = squareImage.pickFromRaster(image, square);
				if (!pick) {
					continue;
				}
			} catch (Exception e) {
				continue;
			}

			// TODO Get rid of undesirable squares
			// (ex. one that has too sharp angle etc.)

			// Convert the image into deviation data.
			deviationData.setRaster(squareImage);

			// Compare with all markers.
			try {
				final ScreenRectangle rectangle = NapUtils.convertToScreenRectangle(square);
				final NapDetectionResult result = matchingStrategy.getBestMatch(
						deviationData, markers, rectangle, getSession());

				// Store the result in the holder.
				if (result != null) {
					resultHolder.push(result);
				}
			} catch (NyARException e) {
				// Do nothing.
			}
		}
		return resultHolder;
	}

	public VariableLengthArray<NapDetectionResult> getLastMarkerDetectionResult() {
		return new VariableLengthArray<NapDetectionResult>(resultHolder);
	}

}
